@@ -1,6 +1,7 @@
 import uuid, json, hashlib, re, time, io, traceback, base64
 from metacat.util import to_bytes, to_str, epoch
 from metacat.auth import BaseDBUser
+from .common import chunked
 from psycopg2 import IntegrityError
 
 Debug = False
@@ -203,7 +204,7 @@ class DBFileSet(object):
 
         file_meta_exp = MetaExpressionDNF(basic_file_query.Wheres).sql(f) or "true"
 
-        datasets = DBDataset.datasets_for_selectors(db, basic_file_query.DatasetSelectors)
+        datasets = None if basic_file_query.DatasetSelectors is None else DBDataset.datasets_for_selectors(db, basic_file_query.DatasetSelectors)
         attrs = DBFile.attr_columns(f)
         if datasets is None:
             # no dataset selection
@@ -305,7 +306,6 @@ class DBFileSet(object):
 
     @staticmethod
     def from_sql(db, sql):
-        print("")
         c = db.cursor()
         debug("DBFileSet.from_sql: executing sql:", sql)
         c.execute(sql)
@@ -313,8 +313,6 @@ class DBFileSet(object):
         fs = DBFileSet.from_tuples(db, fetch_generator(c))
         fs.SQL = sql
         return fs
-    
-
         
 class DBFile(object):
     
@@ -324,9 +322,7 @@ class DBFile(object):
     def __init__(self, db, namespace = None, name = None, metadata = None, fid = None, size=None, checksums=None,
                     parents = None, children = None, creator = None, created_timestamp=None,
                     ):
-        
-        #print("DBFile.__init__: creator=", creator)            
-        
+
         assert (namespace is None) == (name is None)
         self.DB = db
         self.FID = fid or self.generate_id()
@@ -538,7 +534,8 @@ class DBFile(object):
                 name text);
             truncate table {temp_table};
                 """)
-        c.copy_from(io.StringIO(strio.getvalue()), temp_table)
+        cvs = strio.getvalue()
+        c.copy_from(io.StringIO(cvs), temp_table)
         #print("DBFile.get_files: strio:", strio.getvalue())
         
         columns = DBFile.all_columns("f")
@@ -549,6 +546,8 @@ class DBFile(object):
                  where t.id = f.id or f.namespace = t.namespace and f.name = t.name
         """
         
+        #print("   sql:", sql)
+        
         #c.execute(sql)
         #for row in c.fetchall():
         #    print(row)
@@ -652,11 +651,11 @@ class DBFile(object):
             data["parents"] = [f.FID for f in self.parents()]
             data["children"] = [f.FID for f in self.children()]
         if with_datasets:
-            data["datasets"] = ["%s:%s" % tup for ds in self.datasets]
+            data["datasets"] = [{"namespace":ns, "name":n} for ns, n in self.datasets]
         return data
 
-    def to_json(self, with_metadata = False, with_provenance=False):
-        return json.dumps(self.to_jsonable(with_metadata=with_metadata, with_provenance=with_provenance))
+    def to_json(self, with_metadata = False, with_datasets = False, with_provenance=False):
+        return json.dumps(self.to_jsonable(with_metadata=with_metadata, with_provenance=with_provenance, with_datasets=with_datasets))
         
     def children(self, as_files=False, with_metadata = False):
         if self.Children is None:
@@ -725,16 +724,7 @@ class DBFile(object):
     def datasets(self):
         return _DBManyToMany(self.DB, "files_datasets", "dataset_namespace", "dataset_name", file_id = self.FID)
 
-    def __datasets(self):
-        # list all datasets this file is found in
-        c = self.DB.cursor()
-        c.execute("""
-            select fds.dataset_namespace, fds.dataset_name
-                from files_datasets fds
-                where fds.file_id = %s
-                order by fds.dataset_namespace, fds.dataset_name""", (self.FID,))
-        return (DBDataset(self.DB, namespace, name) for namespace, name in fetch_generator(c))
-        
+
 class MetaExpressionDNF(object):
     
     def __init__(self, exp):
@@ -1155,7 +1145,7 @@ class DBDataset(DBObject):
     def add_file(self, f, **args):
         return self.add_files([f], **args)
         
-    def add_files(self, files, do_commit=True, validate_meta=True):
+    def ___add_files(self, files, do_commit=True, validate_meta=True):
         if isinstance(files, DBFile):
             files = [files]
         files = list(files)     # in case it was a generator
@@ -1182,6 +1172,42 @@ class DBDataset(DBObject):
             raise
         return self
 
+    def add_files(self, files, do_commit=True, validate_meta=True):
+        if isinstance(files, DBFile):
+            files = [files]
+        meta_errors = []
+        c = self.DB.cursor()
+        c.execute("begin")
+        t = int(time.time()*1000) % 1000000
+        temp_table = f"temp_{t}"
+        try:
+            c.execute(f"create temp table if not exists {temp_table} (fid text, namespace text, name text)")
+            c.execute(f"truncate table {temp_table}")
+            for chunk in chunked(files, 1000):
+                if validate_meta:
+                    for f in chunk:
+                        assert isinstance(f, DBFile)
+                        errors = self.validate_file_metadata(f.Metadata)
+                        if errors:
+                            meta_errors += errors
+
+                csv = "\n".join(["%s\t%s\t%s" % (f.FID, self.Namespace, self.Name) for f in chunk])
+                c.copy_from(io.StringIO(csv), temp_table, columns = ["fid", "namespace", "name"])
+
+            if meta_errors:
+                raise MetaValidationError("File metadata validation errors", meta_errors)
+
+            c.execute(f"""
+                insert into files_datasets(file_id, dataset_namespace, dataset_name) 
+                    select fid, namespace, name from {temp_table}
+                    on conflict do nothing""")
+            c.execute(f"drop table {temp_table}")
+            c.execute("commit")
+        except:
+            c.execute("rollback")
+            raise
+        return self
+
     def list_files(self, with_metadata=False, limit=None):
         meta = "null as metadata" if not with_metadata else "f.metadata"
         limit = f"limit {limit}" if limit else ""
@@ -1488,17 +1514,54 @@ class DBDataset(DBObject):
         
     @staticmethod
     def datasets_for_files(db, files):
-        file_ids = [f.FID for f in files]
+        #
+        # files: list of DBFile objects. Each DBFile has either valid FID or Namespace, Name
+        #        or dicts:
+        #           {"namespace":"...", "name":"..."}
+        #           {"fid":"..."}
+        #
+        # returns dict:
+        #    { "file_id" -> [DBDataset, ...] }
+        #    for files specified with namespace/name pairs, the output dictionary will also contain
+        #    { ("namespace", "name") -> [DBDataset, ...] }
+        #
+        file_ids = set()
+        pairs = []
+        for f in files:
+            if isinstance(f, DBFile):
+                if f.FID:
+                    file_ids.add(f.FID)
+                else:
+                    pairs.append(dict(namespace=f.Namespace, name=f.Name))
+            elif isinstance(f, dict):
+                fid = f.get("fid")
+                if fid:
+                    file_ids.add(fid)
+                else:
+                    namespace, name = f.get("namespace"), f.get("name")
+                    if not (namespace and name):
+                        raise ValueError("Unrecognozed file specification:", f)
+                    pairs.append((namespace, name))
+            else:
+                raise ValueError("Unrecognozed file specification:", f)
+
+        fid_to_pair = {}
+        if pairs:
+            pair_files = DBFile.get_files(db, pairs)
+            for f in pair_files:
+                file_ids.add(f.FID)
+                fid_to_pair[f.FID] = (f.Namespace, f.Name)
+
         dataset_map = {}       # { fid -> [DBDataset, ...]}
         datasets = {}       # {(ns,n) -> DBDataset}
         c = db.cursor()
         ds_columns = self.columns("ds")
+        
         c.execute(f"""
-            select distinct f.id, {ds_columns}
-                        from datasets ds, files f, files_datasets fd
-                        where f.id = any(%s) and
-                            fd.dataset_namespace = ds.namespace and fd.dataset_name = ds.name and fd.file_id = f.id
-                        order by f.id, ds.namespace, ds.name
+            select distinct fd.file_id, {ds_columns}
+                        from datasets ds, files_datasets fd
+                        where fd.dataset_namespace = ds.namespace and fd.dataset_name = ds.name and fd.file_id = any(%s)
+                        order by fd.file_id, ds.namespace, ds.name
                         """, (file_ids,)
         )
         
@@ -1510,9 +1573,14 @@ class DBDataset(DBObject):
                 ds = datasets[(namespace, name)] = DBDataset.from_tuple(db, tup[1:])
             dslist = dataset_map.setdefault(fid, [])
             dslist.append(ds)
-        
+            
+            ns_n_pair = fid_to_pair.get(fid)
+            if ns_n_pair:
+                dslist = dataset_map.setdefault(ns_n_pair, [])
+                dslist.append(ds)
+
         return dataset_map
-        
+
 class DBNamedQuery(object):
 
     def __init__(self, db, namespace, name, source, parameters=[]):
@@ -1714,7 +1782,6 @@ class DBNamespace(object):
                         from namespaces
             """
             args = ()
-        #print("DBNamespace.list: sql, args:", sql, args)
         c.execute(sql, args)
         for name, owner_user, owner_role, description, creator, created_timestamp in c.fetchall():
             ns = DBNamespace(db, name, owner_user, owner_role, description)
