@@ -27,7 +27,8 @@ class ServerError(Exception):
         self.Body = to_str(body)
         
     def __str__(self):
-        msg = f"MetaCatServer error:\n  URL: {self.URL}\n  HTTP status code: {self.StatusCode}\n  Message: {self.Message}"
+        #msg = f"MetaCatServer error:\n  URL: {self.URL}\n  HTTP status code: {self.StatusCode}\n  {self.Message}"
+        msg = self.Message
         if self.Body:
             msg += "\nMessage from the server:\n"+self.Body+"\n"
         return msg
@@ -74,17 +75,21 @@ class HTTPClient(object):
         self.ServerURL = server_url
         self.Token = token
         self.Timeout = timeout or self.DefaultTimeout
-        
+        self.LastResponse = self.LastURL = self.LastStatusCode = None
+
     def retry_request(self, method, url, timeout=None, **args):
         """
         Implements the functionality to retry on 503 response with random exponentially growing delay
+        Use timemout = 0 to try the request exactly once
+        Returns the response with status=503 on timeout
         """
         if timeout is None:
             timeout = self.DefaultTimeout
         tend = time.time() + timeout
         retry_interval = self.InitialRetry
         response = None
-        while time.time() < tend:
+        done = False
+        while not done:
             if method == "get":
                 response = requests.get(url, timeout=self.Timeout, **args)
             else:
@@ -94,84 +99,52 @@ class HTTPClient(object):
                 break
             sleep_time = min(random.random() * retry_interval, tend-time.time())
             retry_interval *= self.RetryExponent
-            if sleep_time > 0:
+            if sleep_time >= 0:
                 time.sleep(sleep_time)
+            else:
+                break       # time out
         return response
 
-    def get_text(self, uri_suffix):
-        url = "%s/%s" % (self.ServerURL, uri_suffix)
-        headers = {}
+    def send_request(self, method, uri_suffix, headers=None, timeout=None, **args):
+        self.LastURL = url = "%s/%s" % (self.ServerURL, uri_suffix)
+        if headers:
+            headers = headers.copy()
         if self.Token is not None:
+            headers = headers or {}
             headers["X-Authentication-Token"] = self.Token.encode()
-        response = self.retry_request("get", url, headers=headers)
+        self.LastResponse = response = self.retry_request(method, url, headers=headers, **args)
         #print(response, response.text)
+        self.LastStatusCode = response.status_code
         if response.status_code == INVALID_METADATA_ERROR_CODE:
             raise InvalidMetadataError(url, response.status_code, response.text)
         if response.status_code == 404:
             raise NotFoundError(url, response.status_code, response.text)
         elif response.status_code != 200:
             raise WebAPIError(url, response.status_code, response.text)
-        return response.text
+        return response
 
-    def get_json(self, uri_suffix):
-        text = self.get_text(uri_suffix)
-        return json.loads(text)
-        
-    def get_json_stream(self, uri_suffix):
-        url = "%s/%s" % (self.ServerURL, uri_suffix)
-        headers = {"Accept": "application/json-seq"}
-        if self.Token is not None:
-            headers["X-Authentication-Token"] = self.Token.encode()
-
-        with self.retry_request("get", url, headers=headers, stream=True) as response:
-            if response.status_code == INVALID_METADATA_ERROR_CODE:
-                raise InvalidMetadataError(url, response.status_code, response.text)
-            if response.status_code == 404:
-                raise NotFoundError(url, response.status_code, response.text)
-            elif response.status_code != 200:
-                raise WebAPIError(url, response.status_code, response.text)
-            
-            if response.headers.get("Content-Type") != "application/json-seq":
-                raise WebAPIError(url, 200, "Expected content type application/json-seq. Got %s instead." % (response.headers.get("Content-Type"),))
-
-            for line in response.iter_lines():
-                if line:    line = line.strip()
-                while line.startswith(b'\x1E'):
-                    line = line[1:]
-                if line:
-                    #print(f"stream line:[{line}]")
-                    obj = json.loads(line)
-                    yield obj
+    def get_text(self, uri_suffix):
+        return self.send_request("get", uri_suffix).text
 
     def post_text(self, uri_suffix, data):
-        #print("post_json: data:", type(data), data)
-        if isinstance(data, (dict, list)):
-            data = json.dumps(data)
-        else:
-            data = to_bytes(data)
-        #print("post_json: data:", type(data), data)
-            
-        url = "%s/%s" % (self.ServerURL, uri_suffix)
-        
-        headers = {}
-        if self.Token is not None:
-            headers["X-Authentication-Token"] = self.Token.encode()
-        #print("HTTPClient.post_json: url:", url)
-        #print("HTTPClient.post_json: data:", data)
-        
-        response = self.retry_request("post", url, data = data, headers = headers)
-        if response.status_code == INVALID_METADATA_ERROR_CODE:
-            #print("raising InvalidMetadataError")
-            raise InvalidMetadataError(url, response.status_code, response.text)
-        if response.status_code != 200:
-            raise WebAPIError(url, response.status_code, response.text)
-        #print("response.text:", response.text)
-        return response.text
-        
+        return self.send_request("post", uri_suffix, data=data).text
+
+    def unpack_json(self, json_text):
+        results = json.loads(json_text)
+        if isinstance(results, dict):
+            if "results" in results:
+                results = results["results"]
+            elif "error" in results:
+                message =  "Server side error: " + results["error"]["type"] + ": "
+                message += results["error"]["value"]
+                raise ServerError(self.LastURL, self.LastStatusCode, message)
+        return results
+                
+    def get_json(self, uri_suffix):
+        return self.unpack_json(self.send_request("get", uri_suffix).text)
+
     def post_json(self, uri_suffix, data):
-        text = self.post_text(uri_suffix, data)
-        return json.loads(text)
-        
+        return self.unpack_json(self.send_request("post", uri_suffix, data=data).text)
 
 class MetaCatClient(HTTPClient, TokenAuthClientMixin):
     
@@ -283,7 +256,7 @@ class MetaCatClient(HTTPClient, TokenAuthClientMixin):
             return self.get_json(f"data/dataset?dataset={spec}")
         except NotFoundError:
             return None
-            
+
     def get_dataset_files(self, did, namespace=None, name=None, with_metadata=False):
         """Gets single dataset
         
@@ -386,7 +359,7 @@ class MetaCatClient(HTTPClient, TokenAuthClientMixin):
         namespace : str, optional
             Default namespace. If a ``file_list`` item is specified with a name without a namespace, the ``default namespace``
             will be used.
-        
+
         Returns
         -------
         list
@@ -423,7 +396,8 @@ class MetaCatClient(HTTPClient, TokenAuthClientMixin):
 
     def declare_file(self, did=None, namespace=None, name=None, auto_name=None,
                      dataset_did=None, dataset_namespace=None,
-                     dataset_name=None, size=0, metadata={}, fid=None, parents=[], checksums={}):
+                     dataset_name=None, size=0, metadata={}, fid=None, parents=[], checksums={},
+                     dry_run=False):
         """Declare new file and add it to the dataset. Requires client authentication.
         
         Parameters
@@ -448,24 +422,29 @@ class MetaCatClient(HTTPClient, TokenAuthClientMixin):
             file metadata, default empty dictionary
         fid : str
             file id, default None - to be auto-generated
-        parents : list of dicts
-            each dict represents one parent file. The dict may contain the following keys:
-                     "fid" - parent file id
-                     "namespace" and "name" - parent file namespace and name
-                     "did" - parent file DID ("<namespace>:<name>")
         checksums : dict
             dictionary with checksum values by the checksum type: {"type":"value", ...}
-        
+        parents : list of dicts
+            each dict represents one parent file. The dict must contain one of the the following
+                - "fid" - parent file id
+                - "namespace" and "name" - parent file namespace and name
+                - "did" - parent file DID ("<namespace>:<name>")
+        dry_run : boolean
+            If true, run all the necessary checks but stop short of actual file declaraion or adding to a dataset. 
+            If not all checks are successful, generate eirher InvalidMetadataError or WebApiError.
+            Default: False = do declare
+
         Returns
         -------
         dict
-            dictionary with file name, namespace and file id
-                     
+            dictionary with file name, namespace and file id. Names and file ids will be auto-generated as necessary.
+
         Notes
         -----
         At least one of the following must be specified for the file:
             - did
             - namespace and either name or auto_name
+
         At least one of the following must be specified for the dataset:
             - dataset_did
             - dataset_namespace and dataset_name
@@ -505,7 +484,7 @@ class MetaCatClient(HTTPClient, TokenAuthClientMixin):
             info["auto_name"] = auto_name
         return self.declare_files(dataset_did, [info])[0]
 
-    def declare_files(self, dataset, files, namespace=None):
+    def declare_files(self, dataset, files, namespace=None, dry_run=False):
         """Declare new files and add them to an existing dataset. Requires client authentication.
         
         Parameters
@@ -518,7 +497,11 @@ class MetaCatClient(HTTPClient, TokenAuthClientMixin):
             For convenience, if declaring single file, the argument can be the single file dictionary instead of a list.
         namespace: str, optional
             Default namespace for files to be declared
-        
+        dry_run : boolean
+            If true, run all the necessary checks but stop short of actual file declaraion or adding to a dataset. 
+            If not all checks are successful, generate eirher InvalidMetadataError or WebApiError.
+            Default: False = do declare
+
         Returns
         -------
         list
@@ -579,7 +562,7 @@ class MetaCatClient(HTTPClient, TokenAuthClientMixin):
             f["namespace"] = namespace
             size = f.get("size")
             if not isinstance(size, int) or size < 0:
-                raise ValueError("File size is unspecified or invalid for for {namespace}:{name} (#{i} in the list)")
+                raise ValueError(f"File size is unspecified or invalid for file #{i} in the list")
 
             meta = item.get("metadata", {})
             for k in meta.keys():
@@ -590,6 +573,7 @@ class MetaCatClient(HTTPClient, TokenAuthClientMixin):
             lst.append(f)
 
         url = f"data/declare_files?dataset={dataset}"
+        if dry_run: url += "&dry_run=yes"
         out = self.post_json(url, lst)
         return out
 
@@ -763,7 +747,7 @@ class MetaCatClient(HTTPClient, TokenAuthClientMixin):
 
         return self.post_json(url, new_list) 
         
-    def get_file(self, name=None, namespace=None, fid=None, did=None, with_metadata = True, with_provenance=True):
+    def get_file(self, name=None, namespace=None, fid=None, did=None, with_metadata = True, with_provenance=True, with_datasets=False):
         """Get one file record
         
         Parameters
@@ -777,8 +761,10 @@ class MetaCatClient(HTTPClient, TokenAuthClientMixin):
             "nemaspace:name"
         with_metadata : boolean
             whether to include file metadata
-        with_provenance:
+        with_provenance : boolean
             whether to include parents and children list
+        with_datasets : boolean
+            whether to include the list of datasets the file is in
 
         Returns
         -------
@@ -793,12 +779,21 @@ class MetaCatClient(HTTPClient, TokenAuthClientMixin):
                     "creator":  "...",                  # username of the file creator
                     "created_timestamp":   ...,         # numeric UNIX timestamp
                     "size": ...,                        # file size in bytes
+                    "checksums": { ... },               # file checksums
+
+                    # included if with_provenance=True
                     "parents":  ["fid",...],            # list of ids for the file parent files
                     "children": ["fid",...],            # list of ids for the file child files
+
+                    # included if with_metadata=True
                     "metadata": { ... },                # file metadata
-                    "checksums": { ... }                # file checksums
+
+                    # included if with_datasets=True
+                    "datasets": [
+                        {"namespace":"...", "name":"..."}, ...
+                    ]
                 }
-        
+
         Notes
         -----
         Retrieving file provenance and metadata takes slightly longer time
@@ -807,7 +802,8 @@ class MetaCatClient(HTTPClient, TokenAuthClientMixin):
             "Either DID or file id or namespace and name must be specified"
         with_meta = "yes" if with_metadata else "no"
         with_rels = "yes" if with_provenance else "no"
-        url = f"data/file?with_metadata={with_meta}&with_provenance={with_rels}"
+        with_datasets = "yes" if with_datasets else "no"
+        url = f"data/file?with_metadata={with_meta}&with_provenance={with_rels}&with_datasets={with_datasets}"
         if did:
             namespace, name = parse_name(did, None)
             assert namespace is not None, f"Invalid DID format: {did}"
@@ -858,7 +854,7 @@ class MetaCatClient(HTTPClient, TokenAuthClientMixin):
             url += f"&add_to={add_to}"
         results = self.post_json(url, query)
         return results
-        
+
     def async_query(self, query, data=None, **args):
         """Run the query asynchronously. Requires client authentication if save_as or add_to are used.
         
@@ -957,20 +953,39 @@ class MetaCatClient(HTTPClient, TokenAuthClientMixin):
         
         return self.post_json(f"data/namespaces", names)
         
-    def list_namespaces(self, pattern=None):
+    def list_namespaces(self, pattern=None, owner_user=None, owner_role=None, directly=False):
         """Creates new namespace
         
         Parameters
         ----------
         pattern : str
             Optional fnmatch style pattern to filter namespaces by name
+        owner_user : str
+            Optional, return only namespaces owned by the specified user
+        directly : boolean
+            If False and owner_user is specified, return also namespaces owned by all roles the user is in
+            Ignored if owner_user is not specified
+        owner_role : str
+            Optional, return only namespaces owned by the specified role.
+            Ignored if owner_user is also specified
 
         Returns
         -------
         list 
-            List of dictionaries with namespace information
+            List of dictionaries with namespace information sorted by the namespace name
         """
-        lst = self.get_json("data/namespaces")
+        url = "data/namespaces"
+        args = ""
+        if owner_user:
+            args += f"owner_user={owner_user}"
+            if directly:
+                args += "&directly=yes"
+        if owner_role:
+            if args: args += "&"            # low level API on the server side will ignore owner_role if owner_user is present, but pass both anyway
+            args += f"owner_role={owner_role}"
+        if args:
+            args = '?' + args
+        lst = self.get_json("data/namespaces" + args)
         for item in lst:
             if pattern is None or fnmatch.fnmatch(item["name"], pattern):
                 yield item
